/* Corsair - Qtopia RSS/Podcast Aggregator
 *
 * Copyright (C) 2005 Joe Rumsey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <qmainwindow.h>
#include <qpe/qpeapplication.h>

#include "corsair.h"

#include <qtextbrowser.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qfile.h>
#include <qstatusbar.h>
#include <qlineedit.h>
#include <qhbox.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qclipboard.h>
#include <qpushbutton.h>
#include <qmenubar.h>
#include <qprogressbar.h>
#include <qmessagebox.h>

#include <qtopia/resource.h>
#include <qtopia/config.h>
#include <qtopia/global.h>
#include <qtopia/qcopenvelope_qws.h>


#include "download.h"
#include "info.h"

#define TOOLBUTTON(pix, slot) \
	{\
		QToolButton *button = new QToolButton(m_bar);\
		button->setPixmap(Resource::loadPixmap(pix));\
		connect(button, SIGNAL(clicked()), this, SLOT(slot()));\
	}

#define INFOBUTTON(but, text, parent, slot) \
{ \
	but = new QPushButton(text, parent);\
	connect(but, SIGNAL(clicked()), this, SLOT(slot()));\
}

CorsairWindow::CorsairWindow(QPEApplication *app, QWidget *parent, const char *name) :
	QMainWindow(parent, name, WStyle_Customize | WType_TopLevel | WStyle_SysMenu | WStyle_Title | WStyle_NormalBorder),
	m_feeds(this),
	m_showDL(NULL),
	m_infoDialog(NULL),
	m_infoBrowser(NULL),
	m_app(app)
{
	m_bar = new QToolBar(this, NULL);
	setToolBarsMovable(false);
	addToolBar(m_bar, Left);

	m_feedInfoDialog = NULL;

	QStatusBar *status = new QStatusBar(this);
	status->setMinimumWidth(width());
	connect(this, SIGNAL(statusUpdate(const QString &)),
			status, SLOT(message(const QString &)));

	m_progress = new QProgressBar(status, "progress");
	status->addWidget(m_progress, 0, true);

	//m_browser = new QTextBrowser(this, NULL);
	m_list = new QListView(this, NULL);
	m_list->addColumn("Name");

	// If I ever get around to actually making the dates work
	//m_list->addColumn("Updated");

	m_list->setAllColumnsShowFocus(true);

	TOOLBUTTON("details", menu);
	TOOLBUTTON("newfeed", newFeed);
	TOOLBUTTON("editdelete", delSelected);
	TOOLBUTTON("get", grabSelected);
	TOOLBUTTON("UtilsIcon", viewSelected);
	TOOLBUTTON("reload", getAllFeeds);
	TOOLBUTTON("play", playSelected);

	// FIXME - how do some programs put menus and buttons in the titlebar?  Wasting
	// all this space is dumb otherwise.
	// Further note - the only programs that do this seem to be Archos programs.
	// I can't find any standard QT(opia) programs that do it.
	m_menuPopup = new QPopupMenu(this);
	QPopupMenu *file = new QPopupMenu(m_menuPopup);
	file->insertItem("&Load", this, SLOT(loadNewFile()));
	file->insertItem("E&xit", this, SLOT(close()));

	QPopupMenu *view = new QPopupMenu(m_menuPopup);
	view->insertItem("&Downloads", this, SLOT(viewDownloads()));
	view->insertItem("&Preferences", this, SLOT(preferences()));

	QPopupMenu *help = new QPopupMenu(m_menuPopup);
	help->insertItem("&Help", this, SLOT(help()));
	help->insertItem("&About", this, SLOT(about()));
	m_menuPopup->insertItem("&File", file);
	m_menuPopup->insertItem("&View", view);
	m_menuPopup->insertItem("&Help", help);

	//QMenuBar *menu = this->menuBar();
	//menu->insertItem("&File", file);
	//menu->move(menu->x(), menu->y() - 10);

	connect(m_list, SIGNAL(doubleClicked(QListViewItem *)), this, SLOT(doubleClick(QListViewItem *)));
	connect(m_list, SIGNAL(returnPressed(QListViewItem *)), this, SLOT(returnPressed(QListViewItem *)));
	connect(m_list, SIGNAL(expanded(QListViewItem *)), this, SLOT(expand(QListViewItem *)));

	setCentralWidget(m_list);
}

void CorsairWindow::showEvent(QShowEvent *ev)
{
	QMainWindow::showEvent(ev);
	initCorsair();
}

void CorsairWindow::initCorsair()
{
	m_runningDownloads.setAutoDelete(false);
	m_pendingDownloads.setAutoDelete(false);

	checkOldDirectories(topDir(false));

	m_feeds.Load(option("sources", "feeds.opml"));
	
	emit statusUpdate("Ready");

	ShowFeeds();

	connect(m_app, SIGNAL(aboutToQuit()), this, SLOT(cleanup()));
}

CorsairWindow::~CorsairWindow()
{
	LOG(1, "~CorsairWindow()\n");
	cleanup();
}

void CorsairWindow::cleanup()
{
	m_runningDownloads.setAutoDelete(true);
	m_runningDownloads.clear();

	m_pendingDownloads.setAutoDelete(true);
	m_pendingDownloads.clear();
}

void CorsairWindow::ShowFeeds()
{
	m_feeds.addToList(m_list);
}

QString CorsairWindow::homeDir()
{
	QString home = Global::homeDirPath();
	if(home == "/home" && option("homeReallyIsHome", "false") != "true") {
		home = "/media";
	}
	return home;
}


QString CorsairWindow::topDir(bool create)
{
	QString name = homeDir() + "/" + option("topDir", "Corsair");
	if(create) {
		QDir topdir(name);
		if(!topdir.exists()) {
			topdir.mkdir(name);
		}
	}
	return name;
}

QString CorsairWindow::feedDir()
{
	return topDir() + "/" + "feeds";
}

QString CorsairWindow::showsDir()
{
	return topDir() + "/" + "shows";
}

void CorsairWindow::doubleClick(QListViewItem *item)
{
	ITEM_TYPE t = m_feeds.getType(item);
	if(t == IT_FEED) {
		return;
	} else if(t == IT_SHOW) {
		FeedShow *show = m_feeds.GetShow(item);
		if(show && show->m_haveLocalCopy) {
			play(show);
		} else {
			grabItem(item);
		}
	}
}

void CorsairWindow::returnPressed(QListViewItem *item)
{
	view(item);
}

Download *CorsairWindow::startFeedDownload(Feed *f, bool isNew)
{
	if(!isNew && !f->m_loaded) {
		m_feeds.reload(f, QString(""), false);
	}

	QUrl url(f->m_url);

	
	QString fd = feedDir();
	QDir dir(fd);
	if(!dir.exists()) {
		dir.mkdir(fd);
	}

	QString fileName = fd + "/";
	if(f->m_name.isEmpty()) {
		fileName += "newfeed.xml";
	} else {
		fileName += f->fileName();
	}

	Download *dl = new Download(this, url, fileName, f);
	connect(dl, SIGNAL(downloadComplete(Download *, int)),
			f, SLOT(downloadComplete(Download *, int)));
	connect(dl, SIGNAL(downloadComplete(Download *, int)), this, SLOT(downloadComplete(Download *, int)));

	m_runningDownloads.append(dl);
	dl->go();

	return dl;
}

void CorsairWindow::grabItem(QListViewItem *item)
{
	if(!item)
		return;

	ITEM_TYPE t = m_feeds.getType(item);

	QString corsairDir = topDir();
	
	// Don't let device power off while downloading.  Will be reenabled when download completes.
	QPEApplication::setTempScreenSaverMode(QPEApplication::DisableSuspend);

	Download *dl = NULL;

	if(t == IT_FEED) {
		LOG(1, "grab item %s\n", (const char *)item->text(0));
		Feed *f = m_feeds.GetFeed(item);
		dl = startFeedDownload(f, false);
	} else if(t == IT_SHOW) {
		LOG(1, "grab show %s\n", (const char *)item->text(0));
		QString sd = showsDir();
		QDir dir(sd);
		if(!dir.exists()) {
			dir.mkdir(sd);
		}

		FeedShow *s = m_feeds.GetShow(item);

		startShowDownload(s);
	}
}

void CorsairWindow::startShowDownload(FeedShow *s)
{
	bool startShowDL = false;

	Download *pend;
	if(m_showDL == s) {
		LOG(1, "Download of %s already in progress\n", (const char *)s->m_enclosure);
		return;
	}

	for(pend = m_pendingDownloads.first(); pend; pend = m_pendingDownloads.next()) {
		if(pend->param() == s) {
			LOG(1, "Download of %s already pending\n", (const char *)s->m_enclosure);
			return;
		}
	}

	QString subDirName;
	QString fileName = s->fullLocalPath(subDirName);
	
	LOG(1, "subDir=%s\n", (const char *)subDirName);
	
	QDir sub(subDirName);
	if(!sub.exists()) {
		sub.mkdir(subDirName);
	}
	
	QUrl url(s->m_enclosure);
	
	if(!m_showDL) {
		m_showDL = s;
		startShowDL = true;
	}
	
	Download *dl = new Download(this, url, fileName, s);
	connect(dl, SIGNAL(downloadComplete(Download *, int)), this, SLOT(downloadComplete(Download *, int)));
	if(startShowDL) {
		m_runningDownloads.append(dl);
		dl->go();
	} else {
		m_pendingDownloads.append(dl);
	}
}

bool CorsairWindow::startEnclosureDownload(const QString &url)
{
	FeedShow *s = m_feeds.findEnclosureShow(url);
	if(s) {
		startShowDownload(s);
		return true;
	} else {
		return false;
	}
}

void CorsairWindow::newFeed()
{
	LOG(2, "new feed\n");
	QDialog dlg(this, "newFeed", true);
	dlg.setCaption(tr("Add feed"));
	QHBoxLayout l(&dlg);
	l.setAutoAdd(true);
	//QHBox h(&dlg);

	QLabel label("Enter URL:", &dlg);
	// Default text from clipboard.
	QClipboard *cb = QApplication::clipboard();	
	QLineEdit e(cb->text(), &dlg, "B");

	dlg.setMinimumWidth(width()-10);

	int result = dlg.exec();

	if(result && !e.text().isEmpty()) {
		LOG(1, "result=%d, %s\n", result, e.text().latin1());
		Feed *f = m_feeds.addFeed("", e.text());
		ShowFeeds();
		startFeedDownload(f, true);
	}
}

void CorsairWindow::grabSelected()
{
	LOG(1, "grab selected\n");
	grabItem(m_list->selectedItem());
}

void CorsairWindow::downloadComplete(Download *dl, int res)
{
	LOG(2, "downloadComplete(%s)\n", (const char *)dl->fileName());

	bool deleteIt = false;
	if(!m_runningDownloads.remove(dl)) {
		LOG(1, "Didn't find download in list\n");
	} else {
		deleteIt = true;
	}

	if(m_showDL == dl->param()) {
		m_showDL->setHaveLocalCopy(true);

		Download *next = m_pendingDownloads.first();
		if(next) {
			m_pendingDownloads.remove(next);
			m_runningDownloads.append(next);
			m_showDL = (FeedShow *)next->param();
			next->go();
		} else {
			m_showDL = NULL;
		}
	}

	if(m_runningDownloads.isEmpty()) {
		emit statusUpdate(res == Download::Success ? "Ready" : "Error");
		QPEApplication::setTempScreenSaverMode(QPEApplication::Enable);
	}
	
	delete dl;
}

QString CorsairWindow::option(QString name, QString defval)
{
	Config cfg("corsair", Config::User);
	cfg.setGroup("options");
	return cfg.readEntry(name, defval);
}

void CorsairWindow::setOption(QString name, QString val)
{
	Config cfg("corsair", Config::User);
	cfg.setGroup("options");
	cfg.writeEntry(name, val);
}

void CorsairWindow::setStatus(const QString &s)
{
	emit statusUpdate(s);
}

void CorsairWindow::progressUpdate(Download *dl, int read)
{
	if(dl->param() != m_showDL)
		return;

	QString s;
	if(dl->expectedLen()) {
		m_progress->setTotalSteps(dl->expectedLen());
		m_progress->setProgress(read);

		s.sprintf("%dK/%dK of %s", read / 1024, 
				  dl->expectedLen() / 1024, (const char *)m_showDL->m_title);
	} else {
		s.sprintf("%dK of %s", read / 1024, (const char *)m_showDL->m_title);
	}
	emit statusUpdate(s);

}

void CorsairWindow::viewSelected()
{
	QListViewItem *item = m_list->selectedItem();
	if(!item)
		return;

	view(item);
}

void CorsairWindow::view(QListViewItem *item)
{
	ITEM_TYPE t = m_feeds.getType(item);

	switch(t) {
		case IT_FEED:
		{
			// FIXME - create a dialog for this.  Right now just download.
			viewFeedInfo(item);
			return;
		}
		case IT_SHOW:
		{
			FeedShow *show = m_feeds.GetShow(item);
			if(!show)
				return;

			if(!m_infoDialog) {
				m_infoDialog = new QWidget(NULL, "ShowInfo", WStyle_Customize | WStyle_DialogBorder | WType_Popup | WDestructiveClose);
				QVBoxLayout *l = new QVBoxLayout(m_infoDialog);
				l->setAutoAdd(true);
				m_infoDialog->setMinimumWidth(width());
				m_infoDialog->setMinimumHeight(height() - 20);

				QHBox *box = new QHBox(m_infoDialog, "buttonbox");

				QPushButton *b;

				INFOBUTTON(b, tr("Prev"), box, infoPrev);
				INFOBUTTON(b, tr("Download"), box, infoDownload);
				m_infoActionButton = b;
				INFOBUTTON(b, tr("Browse"), box, infoBrowse);
				INFOBUTTON(m_infoNextButton, tr("Next"), box, infoNext);

				connect(m_infoDialog, SIGNAL(destroyed()), this, SLOT(infoClosed()));

				m_infoBrowser = new InfoBrowser(this, m_infoDialog);
			}

			m_infoNextButton->setFocus();
			m_infoItem = item;

			m_infoDialog->setCaption(show->m_title);
			if(show->m_haveLocalCopy) {
				m_infoActionButton->setText(tr("Play"));
			} else {
				m_infoActionButton->setText(tr("Download"));
			}

			QString text = "<html>" 
				"<H4><a href=\"" + show->m_link + "\">" + show->m_title + "</a></H4><br>" +
				show->m_description + "</html>";

			m_infoBrowser->setText(text);

			LOG(3, "browse text:\n%s\n", (const char *)text);

			show->m_isNew = false;

			m_infoDialog->showMaximized();
			break;
		}
	}
}

void CorsairWindow::delSelected()
{
	QListViewItem *item = m_list->selectedItem();
	if(!item)
		return;

	ITEM_TYPE t = m_feeds.getType(item);

	switch(t) {
		case IT_FEED:
		{
			Feed *feed = m_feeds.GetFeed(item);
			if(!feed)
				return;

			switch(QMessageBox::warning(this, "Confirm delete",
										"Confirm deletion of \n" +
										(feed->m_name.length() > 0 ? feed->m_name : feed->m_url) + "\n",
										"Delete", "Cancel", QString::null, 0, 1)) {
				case 1:
					break;
				case 0:
					m_feeds.delFeed(item);
					break;
			}
			break;
		}
		case IT_SHOW:
			// FIXME (delete file, not entry)
			break;
	}
}

void CorsairWindow::getAllFeeds()
{
	Feed *i;
	for(i = m_feeds.feedList()->first(); i; i = m_feeds.feedList()->next()) {
		startFeedDownload(i, false);
	}
}

void CorsairWindow::infoPrev()
{
	if(m_infoItem->itemAbove() && m_infoItem->itemAbove()->parent() == m_infoItem->parent()) {
		view(m_infoItem->itemAbove());
	}
}

void CorsairWindow::infoDownload()
{
	FeedShow *show = m_feeds.GetShow(m_infoItem);
	if(show && show->m_haveLocalCopy) {
		play(show);
	} else {
		grabItem(m_infoItem);
	}
}

void CorsairWindow::infoNext()
{
	if(m_infoItem->itemBelow() && m_infoItem->itemBelow()->parent() == m_infoItem->parent()) {
		view(m_infoItem->itemBelow());
	}
}

void CorsairWindow::infoBrowse()
{
	FeedShow *s = m_feeds.GetShow(m_infoItem);
	if(s && !s->m_link.isEmpty()) {
		m_infoBrowser->setSource(s->m_link);
	}
}

void CorsairWindow::infoClosed()
{
	LOG(1, "Info dialog closed\n");
	m_infoDialog = NULL;
}

void CorsairWindow::about()
{
	QDialog dlg(this, "about", true);
	dlg.setCaption(tr("About Corsair"));
	QVBoxLayout l(&dlg);
	l.setAutoAdd(true);

	QLabel name("Corsair v"+ version(), &dlg);
	QLabel copyright("Copyright (C) 2005 Joe Rumsey <joe@rumsey.org>", &dlg);
	QLabel homepage("Home page: http://corsair.sourceforge.net/", &dlg);
	dlg.exec();
}

void CorsairWindow::help()
{
	QCString browser = (const char *)QString("QPE/Application/helpbrowser");
	QCopEnvelope e(browser, "setDocument(QString)");
	QString url = "file://" + QPEApplication::qpeDir() + "help/html/corsair.html";
	e << url;
	
	QCopEnvelope r(browser, "raise()");
}

void CorsairWindow::playSelected()
{
	QListViewItem *item = m_list->selectedItem();
	if(!item)
		return;

	FeedShow *show = m_feeds.GetShow(item);
	if(!show)
		return;

	play(show);
}

void CorsairWindow::play(FeedShow *show)
{
	if(show->m_haveLocalCopy) {
		QCString player = (const char *)QString("QPE/Application/" + option("audioplayer", "player"));
		QCopEnvelope e(player, "setDocument(QString)");
		QString sub;
		e << show->fullLocalPath(sub);

		QCopEnvelope r(player, "raise()");
	}
}

void CorsairWindow::menu()
{
	m_menuPopup->exec(QCursor::pos());
}

void CorsairWindow::loadNewFile()
{
}

void CorsairWindow::viewDownloads()
{
}

void CorsairWindow::preferences()
{
}

void CorsairWindow::expand(QListViewItem *item)
{
	Feed *feed = m_feeds.GetFeed(item);
	if(feed && !feed->m_loaded) {
		m_feeds.reload(feed, "", false);
	}
}
		
void CorsairWindow::viewFeedInfo(QListViewItem *item)
{
	Feed *f = m_feeds.GetFeed(item);
	if(!f)
		return;

	if(!f->m_loaded) {
		m_feeds.reload(f, "", false);
	}

	if(!m_feedInfoDialog) {
		m_feedInfoDialog = new FeedInfo(this);
		connect(m_feedInfoDialog, SIGNAL(destroyed()), this, SLOT(feedInfoDestroyed()));
	}
	m_feedInfoDialog->setFeed(f);
	m_feedInfoDialog->showMaximized();
}

void CorsairWindow::feedInfoDestroyed()
{
	m_feedInfoDialog = NULL;
}

void CorsairWindow::saveFeeds()
{
	m_feeds.save();
}

void CorsairWindow::checkOldDirectories(QString newdir)
{
	QString olddir = homeDir() + "/" + "Documents/corsair";
	QDir o(olddir);
	QDir n(newdir);
	if(o.exists() && !n.exists()) {

		QDialog *dlg = new QDialog(this, "Move", true);
		QVBoxLayout *l = new QVBoxLayout(dlg);
		l->setAutoAdd(true);

		QLabel *label = new QLabel("Old directory structure detected, move \n" + olddir + "\n to \n" + newdir + "?", dlg);
		
		QHBox *h = new QHBox(dlg);
		QPushButton *b = new QPushButton("Move it", h, "move");
		connect(b, SIGNAL(clicked()), dlg, SLOT(accept()));
		
		b = new QPushButton("Leave it", h, "leave");
		connect(b, SIGNAL(clicked()), dlg, SLOT(reject()));

		if(dlg->exec()) {
			QDir d;
			d.rename(olddir, newdir);
		}
	}
}
