/* Corsair - Qtopia RSS/Podcast Aggregator
 *
 * Copyright (C) 2005 Joe Rumsey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "corsair.h"

#include "download.h"

Download::Download(CorsairWindow *parent, QUrl &url, QString &filename, void *param) :
	m_fileName(filename),
	m_file(filename)
{
	m_parent = parent;
	m_url = url;
	m_param = param;
	m_complete = false;

	init();

	connect(this, SIGNAL(readyRead()), this, SLOT(readyRead()));
	connect(this, SIGNAL(error(int)), this, SLOT(error(int)));
	connect(this, SIGNAL(hostFound()), this, SLOT(hostFound()));
	connect(this, SIGNAL(connected()), this, SLOT(connected()));
	connect(this, SIGNAL(connectionClosed()), this, SLOT(connectionClosed()));

	connect(this, SIGNAL(statusUpdate(const QString &)), m_parent, SLOT(setStatus(const QString &)));
	connect(this, SIGNAL(progressUpdate(Download *, int)), m_parent, SLOT(progressUpdate(Download *, int)));

}

Download::~Download()
{
	LOG(1, "~Download()\n");
	m_file.close();
	close();
}

void Download::init()
{
	m_headersComplete = false;
	m_chunked = false;
	m_totalRead = 0;
	m_expectedLen = 0;
	m_gotResponse = false;
	m_redirect = false;
	m_binary = false;
}

void Download::go()
{
	int port = m_url.port();
	if(port < 0)
		port = 80;

	LOG(2, "connecting to %s:%d\n", (const char *)m_url.host(), port);
	connectToHost(m_url.host(), port);

	emit statusUpdate("Connecting to " + m_url.host());
}

void Download::openFile()
{
	if(!m_file.isOpen()) {
		LOG(2, "Opening %s for writing\n", (const char *)m_fileName);
		if(!m_file.open(IO_WriteOnly)) {
			LOG(0, "Couldn't open file %s\n", (const char *)m_fileName);
		}
	}
}

void Download::readyRead()
{
	if(m_complete)
		return;

	if(m_binary && m_headersComplete) {
		while(bytesAvailable()) {
			char buf[1024];
			int r = readBlock(buf, sizeof(buf));
			if(r < 0) {
				m_file.close();
				complete(m_totalRead > 0 ? Success : NoData);
			} else {
				m_file.writeBlock(buf, r);
				m_totalRead += r;
				if(m_totalRead >= m_expectedLen) {
					emit progressUpdate(this, m_totalRead);
					close();
					m_file.close();
					complete(Success);
					return;
				}
			}
		}
		emit progressUpdate(this, m_totalRead);
		return;
	}

	while(canReadLine()) {
		LOG(4, "canReadLine\n");
		QString line = readLine();
		if(m_headersComplete) {
			LOG(4, "%s", (const char *)line);
			if(m_chunked) {
				if(m_chunkRemaining <= 0) {
					bool ok;
					m_chunkRemaining = line.toInt(&ok, 16);
					if(!ok) {
						LOG(1, "Bad chunk line %s\n", (const char *)line);
					} else {
						LOG(1, "Chunk: %d\n", m_chunkRemaining);
						if(m_chunkRemaining <= 0) {
							close();
							m_file.close();
							complete(Success);
							return;
						}
					}
					continue;
				} else {
					m_chunkRemaining -= line.length();
					if(m_chunkRemaining < 0) {
						// Chunk marker is always on a new line, but
						// that newline shouldn't appear in the document
						line = line.left(line.length() + m_chunkRemaining);
					}
				}
			}
			m_file.writeBlock((const char *)line, line.length());
			m_totalRead += line.length();
		} else if(!m_gotResponse) {
			LOG(2, "%s", (const char *)line);
			if(line.contains("302 Found") || line.contains("301 Moved Permanently")) {
				m_redirect = true;
			} else if(!line.contains("200 OK")) {
				complete(Error);
				return;
			}
			m_gotResponse = true;
		} else {
			LOG(2, "%s", (const char *)line);
			QString lower = line.lower();
			if(lower.startsWith("transfer-encoding: ") && line.contains("chunked")) {
				m_chunked = true;
				m_chunkRemaining = 0;
			} else if(lower.startsWith("content-type: ") && line.contains("audio")) {
				m_binary = true;
			} else if(lower.startsWith("content-length: ")) {
				m_expectedLen = line.right(line.length() - (QString("content-length: ").length())).toInt();
			} else if(m_redirect && lower.startsWith("location: ")) {
				m_redirectLocation = line.right(line.length() - (QString("location: ").length())).stripWhiteSpace();
			} else if(line.length() < 1 || line[0] == '\r' || line[0] == '\n') {
				if(m_redirect) {
					init();
					close();
					m_url = QUrl(m_redirectLocation);
					go();
				} else {
					m_headersComplete = true;
					openFile();

					if(!m_chunked)
						m_binary = true;
				}
			}
		}
	}
	if(m_headersComplete) {
		emit progressUpdate(this, m_totalRead);
	}
}

void Download::error(int e)
{
	LOG(2, "error %d\n", e);
	m_file.close();
	complete(Error);
}

void Download::connected()
{
	emit statusUpdate("Connected to " + m_url.host());
	LOG(2, "connected\n");

	QString get = m_url.encodedPathAndQuery();

	QString req;
	req = "GET " + get + " HTTP/1.1\r\n";
	req += "Host: " + m_url.host() + "\r\n";
	req += "\r\n";

	LOG(2, "Grabbing %s with '%s'\n", (const char *)get, (const char *)req);

	writeBlock(req, req.length());
	flush();
}

void Download::hostFound()
{
	emit statusUpdate("Host " + m_url.host() + " found");
	LOG(2, "hostFound\n");
}

void Download::connectionClosed()
{
	readyRead();

	emit statusUpdate(m_url.host() + "Done");
	LOG(2, "connectionClosed\n");
	m_file.close();
	complete(Error);
}

void Download::complete(Result res)
{
	LOG(1, "Download::complete(%d)\n", res);
	if(m_complete)
		return;

	m_complete = true;
	emit downloadComplete(this, res);
}

