/* Corsair - Qtopia RSS/Podcast Aggregator
 *
 * Copyright (C) 2005 Joe Rumsey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "corsair.h"

#include "feeds.h"
#include <qxml.h>
#include <qtopia/resource.h>
#include <qdir.h>
#include <qstring.h>


#include "opmlparser.h"
#include "feedparser.h"
#include <ctype.h>

static QPixmap 
    *s_feedIcon = NULL,
	*s_showIcon = NULL,
	*s_oldShowIcon = NULL,
	*s_noEnclosureIcon = NULL,
	*s_groupIcon = NULL;

static QFont
    *s_boldFont = NULL;

class CorsairListViewItem : public QListViewItem
{
private:
	ITEM_TYPE m_type;

protected:
	QPixmap *m_pix;

public:
	CorsairListViewItem(ITEM_TYPE type, QListView *l, QPixmap *pix) :
		QListViewItem(l), m_type(type), m_pix(pix) {
	}

	CorsairListViewItem(ITEM_TYPE type, QListViewItem *i, QPixmap *pix) :
		QListViewItem(i), m_type(type), m_pix(pix) {
	}

	ITEM_TYPE type() { return m_type; }

	const QPixmap *pixmap(int i) const {
		if(i) return NULL;
		return m_pix;
	}

	void setPixmap(QPixmap *p) {
		m_pix = p;
	}
};

class FeedListViewItem : public CorsairListViewItem
{
private:
	Feed *m_feed;

	friend class FeedList;
	friend class Feed;

public:
	FeedListViewItem(Feed *feed, QListViewItem *l, QPixmap *pix) : 
		CorsairListViewItem(IT_FEED, l, pix), m_feed(feed)
	{
	}

	void paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align) {
		FeedShow *s;
		for(s = m_feed->m_shows.first(); s; s = m_feed->m_shows.next()) {
			if(s->m_isNew) {
				p->setFont(*s_boldFont);
				break;
			}
		}
		CorsairListViewItem::paintCell(p, cg, column, width, align);
	}

	QString text(int column) const {
		switch(column) {
			case NAME_COLUMN:
				if(m_feed->m_name.isEmpty()) {
					return m_feed->m_url;
				} else {
					return m_feed->m_name;
				}				
			case DATE_COLUMN:
			{
				QDate d = m_feed->m_lastGrab.date();
				QTime t = m_feed->m_lastGrab.time();
      
				QDateTime now = QDateTime::currentDateTime();

				if(d.day() == now.date().day() &&
				   d.month() == now.date().month() &&
				   d.year() == now.date().year()) {
					return t.toString();
				} else {
					return d.toString();
				}
			}
			default:
				return "";
		}
	}

	void setup() {
		setOpen(false);
		setExpandable(true);
		QListViewItem::setup();
	}
};

class ShowListViewItem : public CorsairListViewItem
{
private:
	FeedShow *m_show;
	friend class FeedList;

public:
	ShowListViewItem(FeedShow *show, FeedListViewItem *parent, QPixmap *pix) :
		CorsairListViewItem(IT_SHOW, parent, pix),
		m_show(show) {
		show->m_listItem = this;
	}

	QString text(int column) const {
		switch(column) {
			case NAME_COLUMN:
				return m_show->m_title;
			case DATE_COLUMN:
				return "";
			default:
				return "";
		}
	}

	QString key(int column, bool ascending) const {
		// Sort shows by their order in the feed, not their name
		if(column == NAME_COLUMN) {
			QString s;
			s.sprintf("%08d", m_show->m_index);
			return s;
		} else {
			return text(column);
		}
	}

	void paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align) {
		if(m_show->m_isNew) {
			p->setFont(*s_boldFont);
		}
		CorsairListViewItem::paintCell(p, cg, column, width, align);
	}

	int width(const QFontMetrics &fm, const QListView *lv, int c) {
		return fm.width(text(c));
	}

};
	
class GroupListViewItem : public CorsairListViewItem
{
private:
	QString m_name;

public:
	GroupListViewItem(QString name, QListView *l, QPixmap *pix) :
		CorsairListViewItem(IT_GROUP, l, pix),
		m_name(name) {
	}

	QString text(int column) const {
		switch(column) {
			case NAME_COLUMN:
				return m_name;
			default:
				return "";
		}
	}
};

FeedShow::FeedShow(Feed *feed,
				   const QString &title,
				   const QString &enclosure,
				   const QString &description,
				   const QString &link,
				   bool isNew,
				   int index) :
	m_feed(feed),
	m_title(title),
	m_enclosure(enclosure),
	m_description(description),
	m_link(link),
	m_isNew(isNew),
	m_index(index),
	m_haveLocalCopy(false) ,
	m_listItem(NULL)
{
	QString sub;
	QFile f(fullLocalPath(sub));
	if(enclosure.length() > 0 && f.exists())
		m_haveLocalCopy = true;
}

QString FeedShow::fullLocalPath(QString &sub)
{
	sub = m_feed->m_list->owner()->showsDir() + "/" + m_feed->dirName() + "/";

	return sub + fileName();
}

void FeedShow::setHaveLocalCopy(bool haveit)
{
	m_haveLocalCopy = haveit;
	if(m_listItem) {
		if(m_haveLocalCopy) {
			m_listItem->setPixmap(s_showIcon);
		} else {
			m_listItem->setPixmap(s_oldShowIcon);
		}
		m_listItem->repaint();
	}
}

	
Feed::Feed(FeedList *list, const QString &name, const QString &url)
{
	LOG(1, "Created feed %s, %s without XmlAttributes", (const char *)name, (const char *)url);
	m_list = list;
	m_name = name;
	m_url = url;
	m_feedItem = NULL;
	m_lastGrab = QDateTime::currentDateTime();
	m_showIndex = 0;
	m_loaded = false;
	m_brokenXML = false;
}

Feed::Feed(FeedList *list, const QXmlAttributes &attr)
{
	m_list = list;
	m_attr = attr;
	m_name = m_attr.value("text");
	m_url = m_attr.value("xmlUrl");
	m_feedItem = NULL;
	m_lastGrab = QDateTime::currentDateTime();
	m_showIndex = 0;
	m_loaded = false;
	m_brokenXML = m_attr.value("brokenXML") == "true";

	//LOG(1, "Created feed %s, %s", (const char *)m_name, (const char *)m_url);
	//Load("");
}

Feed::~Feed()
{
	if(m_feedItem) {
		m_feedItem->m_feed = NULL;
		delete m_feedItem;
	}

	m_shows.setAutoDelete(true);
	m_shows.clear();
}

QString Feed::Link()
{
	return "<a href=\"" + m_url + "\">" + m_name + "</a><br>\n";
}

void Feed::addShow(const QString &title, const QString &enclosure, const QString &description, const QString &link)
{
	FeedShow *i;
	for(i = m_shows.first(); i; i = m_shows.next()) {
		if(link.length() > 0) {
			if(i->m_link == link)
				return;
		} else if(enclosure.length() > 0) {
			if(i->m_enclosure == enclosure) {
				return;
			}
		} else {
			if(i->m_title == title) {
				return;
			}
		}
	}

	LOG(3, "addShow %s, %s, %d\n", (const char *)title, (const char *)enclosure, m_showIndex + 1);

	FeedShow *newshow;
	if(m_list->showsAreNew()) {
		// Put new show before first old show
		FeedShow *s;
		int index = -1;
		for(s = m_shows.first(); s; s = m_shows.next()) {
			if(!s->m_isNew) {
				if(index < 0) {
					index = s->getIndex();
				}
				s->setIndex(s->getIndex() + 1);
			}
		}
		if(index < 0) {
			index = ++m_showIndex;
		}
		newshow = new FeedShow(this, title, enclosure, description, link,
							   m_list->showsAreNew(),
							   index);
	} else {
		newshow = new FeedShow(this, title, enclosure, description, link,
							   m_list->showsAreNew(),
							   ++m_showIndex);
	}
	m_shows.append(newshow);
}

QPixmap *Feed::chooseIcon(FeedShow *show)
{
	if(show->m_enclosure.length() <= 0) {
		return s_noEnclosureIcon;
	} else if(show->m_haveLocalCopy) {
		return s_showIcon;
	} else {
		return s_oldShowIcon;
	}
}

void Feed::updateListItem(QListViewItem *g)
{
	if(!g)
		g = m_list->groupItem();

	if(!m_feedItem) {
		m_feedItem = new FeedListViewItem(this, g, s_feedIcon);
	} else {
		clear();
	}

	FeedShow *i;
	for(i = m_shows.first(); i; i = m_shows.next()) {
		ShowListViewItem *showitem = new ShowListViewItem(i, m_feedItem, chooseIcon(i));
	}
	m_feedItem->repaint();
}

// Load a local cached copy if one exists (not download)
void Feed::Load(QString filename)
{
	FeedParser parser(this);

	// If filename is set, loading from a temp file, rename it at the end
	bool rename = !filename.isEmpty();
	if(filename.isEmpty()) {
		filename = m_list->owner()->feedDir() + "/" + fileName();
	}
	LOG(1, "Loading feed from %s\n", (const char *)filename);

	QFile f(filename);
	if(!f.exists())
		return;

	QString s;
	f.open(IO_ReadOnly);
	s = (const char *)f.readAll();
	if(m_brokenXML) {
		LOG(1, "Feed has broken XML, fixing\n");
		int c;
		int state = 0;
		for(c = 0; c < s.length(); c++) {
			switch(state) {
				case 0:
					if(s[c] == '<')
						state = 1;
					break;
				case 1:
					if(s[c] == '&' && s[c+1] == ' ') {
						s[c] = '+';
					} else if(s[c] == '>') {
						state = 0;
					}
					break;
			}
		}
		f.close();
		f.open(IO_WriteOnly);
		f.writeBlock((const char *)s, s.length());
	}
	f.close();

	//QTextStream stream(&s, IO_ReadOnly);
	QXmlInputSource source(f);
	QXmlSimpleReader reader;
	reader.setContentHandler(&parser);
	reader.setErrorHandler(&parser);
	reader.parse(source);

	if(rename) {
		if(!m_name.isEmpty()) {
			LOG(2, "renaming %s to %s\n", (const char *)filename, (const char *)(m_list->owner()->feedDir() + "/" + fileName()));
			// Loaded from temp file, parsed a name, rename the temp file.
			QDir dir;
			dir.rename(filename, m_list->owner()->feedDir() + "/" + fileName(), true);
		} else {
			LOG(2, "rename was set, but no name was parsed\n");
		}
	}

	m_loaded = true;
}

void Feed::clear()
{
	if(!m_feedItem)
		return;

	while(m_feedItem->firstChild()) {
		delete m_feedItem->firstChild();
	}
}

void Feed::setName(const QString &name)
{ 
	// Only set it if we don't have a name already
	if(m_name.length() <= 0) {
		m_name = name; 
		m_list->save();
	}
}

void Feed::rename(const QString &newname)
{
	QString oldname = m_list->owner()->feedDir() + "/" + fileName();
	QString oldShowsDir = m_list->owner()->showsDir() + "/" + dirName();

	m_name = newname;

	// Rename shows dir first, so we can check for local files while reparsing shows
	QString newShowsDir = m_list->owner()->showsDir() + "/" + dirName();
	QDir dir;
	dir.rename(oldShowsDir, newShowsDir, true);

	// This will rename the feed file too!
	Load(oldname);

}

void Feed::downloadComplete(Download *dl, int result)
{
	LOG(1, "Feed %s download complete\n", (const char *)m_name);
	// Signalled by Download
	QString ff = dl->fileName();
	m_list->reload(this, ff, true);
}

QString Feed::dirName()
{
	QString str = m_name;
	int i;
	for(i = 0; i < str.length(); i++) {
		// Draconian character replacement to make sure the filename
		// (which user should never see from inside the app anyway)
		// is ok.
		char c = (char)((const QString &)str)[i];
		if(!isalnum(c) && str[i] != ' ' && str[i] != '_') {
			str[i] = '_';
		}
	}

	return str;
}

FeedList::FeedList(CorsairWindow *owner)
{
	if(!s_feedIcon) {
		s_feedIcon = new QPixmap(Resource::loadPixmap("list"));
		s_showIcon = new QPixmap(Resource::loadPixmap("volume"));
		s_oldShowIcon = new QPixmap(Resource::loadPixmap("mute"));
		s_groupIcon = new QPixmap(Resource::loadPixmap("fileopen"));
		s_noEnclosureIcon = new QPixmap(Resource::loadPixmap("txt"));
		s_boldFont = new QFont;
		s_boldFont->setBold(true);
	}

	m_owner = owner;
	m_showsAreNew = false;
	m_groupItem = NULL;
}

FeedList::~FeedList()
{
}

void FeedList::Clear()
{
	m_feeds.setAutoDelete(true);
	m_feeds.clear();
}

void FeedList::Load(const QString &justfile)
{
	m_showsAreNew = false;

	QString file = m_owner->topDir() + "/" + justfile;
	m_filename = justfile;

	LOG(1, "Loading %s\n", file.latin1());

	Clear();

	OPMLParser parser(this);

	QFile f(file);
	if(!f.exists()) {
		LOG(1, "file %s does not exist\n", file.latin1());
		addFeed("Corsair News", "http://sourceforge.net/export/rss2_projnews.php?group_id=147195&amp;rss_fulltext=1");
		addFeed("Corsair Releases", "http://sourceforge.net/export/rss2_projfiles.php?group_id=147195");
		save();
	} else {
		QXmlInputSource source(f);
		QXmlSimpleReader reader;

		reader.setContentHandler(&parser);
		reader.setErrorHandler(&parser);

		reader.parse(source);
		
		m_showsAreNew = true;
	}
}

Feed *FeedList::addFeed(const QString &name, const QString &url)
{
	Feed *f = new Feed(this, name, url);
	m_feeds.append(f);
	return f;
}

void FeedList::AddFeed(const QXmlAttributes &attr)
{
	if(attr.index("text") >= 0 && attr.index("xmlUrl") >= 0) {
		m_feeds.append(new Feed(this, attr));
	}
}


void FeedList::addToList(QListView *l)
{
	if(!m_groupItem) {
		m_groupItem = new GroupListViewItem(m_filename, l, s_groupIcon);
		m_groupItem->setOpen(true);
	}

	Feed *f;
	for(f = m_feeds.first(); f; f = m_feeds.next()) {
		f->updateListItem(m_groupItem);
	}
}

void FeedList::reload(Feed *f, const QString &filename, bool showsAreNew)
{
	bool oldShowsAreNew = m_showsAreNew;
	m_showsAreNew = showsAreNew;
	f->Load(filename);
	f->updateListItem(m_groupItem);
	m_showsAreNew = oldShowsAreNew;
}

Feed *FeedList::GetFeed(QListViewItem *item)
{
	if(getType(item) != IT_FEED)
		return NULL;

	FeedListViewItem *fi = (FeedListViewItem *)item;
	return fi->m_feed;
}

FeedShow *FeedList::GetShow(QListViewItem *item)
{
	if(getType(item) != IT_SHOW)
		return NULL;
	
	ShowListViewItem *si = (ShowListViewItem *)item;
	return si->m_show;
}

ITEM_TYPE FeedList::getType(QListViewItem *item)
{
	CorsairListViewItem *ci = (CorsairListViewItem *)item;
	return ci->type();
}

static QString fixForXML(const QString &src)
{
	// Convert any already converted &'s back to plain &s,
	// then convert all of them to &amp;'s
	QString dest = src;
	dest.replace(QRegExp("&amp;"), "&");
	dest.replace(QRegExp("&"), "&amp;");
	return dest;
}

void FeedList::save(QString filename)
{
	if(filename.isEmpty()) {
		filename = m_owner->topDir() + "/" + m_filename;
	}

	LOG(1, "Saving feeds to %s\n", (const char *)filename);

	QFile file(filename);
	if(!file.open(IO_WriteOnly)) {
		LOG(1, "Can't open file %s\n", (const char *)filename);
		return;
	}
	QTextStream f(&file);
	f << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n";
	f << "<!-- OPML generated by Corsair -->\n";
	f << "<opml version=\"1.1\">\n";
	f << "	<head>\n";
	f << "		<title>Corsair Feeds</title>\n";
	f << "	</head>\n";
	f << "	<body>\n";

	Feed *feed;
	for(feed = m_feeds.first(); feed; feed = m_feeds.next()) {
		QString fixedURL = fixForXML(feed->m_url);
		QString fixedName = fixForXML(feed->m_name);

		f << "		<outline text=\"" << fixedName << "\" "
		  << " brokenXML=\"" << (feed->m_brokenXML ? "true" : "false") << "\" "
		  << " xmlUrl=\"" << fixedURL << "\"/>\n";
	}
	f << "	</body>\n";
	f << "</opml>\n";
}

void FeedList::delFeed(QListViewItem *item)
{
	Feed *f = GetFeed(item);
	if(!f)
		return;

	if(m_feeds.remove(f)) {
		save();
	}
}

FeedShow *FeedList::findEnclosureShow(const QString &url)
{
	Feed *f;
	for(f = m_feeds.first(); f; f = m_feeds.next()) {
		FeedShow *s;
		for(s = f->m_shows.first(); s; s = f->m_shows.next()) {
			if(s->m_enclosure == url)
				return s;
		}
	}
	return NULL;
}
